-- Reduction pairs based on linear polynomial interpretations:

module EncodingLinear where

import Term
import Rule
import Order
import SMT
import qualified Linear
import ReductionPair (ReductionPair (..))
import TRS

type Affine = (Exp, [Exp])

-- Encoding functions for linear interpreations.
-- k means the k-th order in lexicographic combination.

-- f_A(x_0,...,x_{n-1}) = a0 + ai_0 x0 + ... + ai_{n-1} x_{n-1}
--   where a0 in N and ai_0, ..., an in {0,1}.

var_a0 :: Int -> String -> Exp
var_a0 k f = Var ("a0_" ++ show k ++ "_" ++ f) 

var_ai :: Int -> String -> Int -> Formula
var_ai k f i = FVar ("ai_" ++ show k ++ "_" ++ show i ++ "_" ++ f) 

constant :: Int -> Term -> Exp
constant _ (V _)    = Val 0
constant k (F f ts) =
  plus (var_a0 k f : 
        [ times01 (var_ai k f i) (constant k ti)
        | (i, ti) <- zip [0..] ts ])

coefficient :: Int -> String -> Term -> Exp
coefficient _ x (V y)
  | x == y    = Val 1
  | otherwise = Val 0
coefficient k x (F f ts) =
  plus [ times01 (var_ai k f i) (coefficient k x ti) 
       | (i, ti) <- zip [0..] ts ]

interpret :: Int -> [String] -> Term -> Affine
interpret k xs t =
  (constant k t, [ coefficient k x t | x <- xs ])

geq_affine :: Affine -> Affine -> Formula
geq_affine (a0, as) (b0, bs) =
  conj (geq a0 b0 : [ geq ai bi | (ai, bi) <- zip as bs ])

gt_affine :: Affine -> Affine -> Formula
gt_affine (a0, as) (b0, bs) =
  conj (gt a0 b0 : [ geq ai bi | (ai, bi) <- zip as bs ])

-- s >=_A t
geq_A :: Int -> Rule -> Formula
geq_A k (s, t) =
  geq_affine (interpret k xs s) (interpret k xs t)
  where xs = Rule.variables (s, t)

-- s >_A t
gt_A :: Int -> Rule -> Formula
gt_A k (s, t) =
  gt_affine (interpret k xs s) (interpret k xs t)
  where xs = Rule.variables (s, t)

side_condition :: Int -> (TRS, TRS) -> Signature -> Formula
side_condition k _ sig =
  conj [ geq (var_a0 k f) (Val 0) | (f, _) <- sig ]

f_i_monotone :: Int -> (String, Int) -> Int -> Formula
f_i_monotone k (f, _) i = var_ai k f i

f_i_invariant :: Int -> (String, Int) -> Int -> Formula
f_i_invariant k (f, _) i = neg (var_ai k f i)

monotone :: Int -> Signature -> Formula
monotone k sig =
  conj [ f_i_monotone k (f, n) i 
       | (f, n) <- sig, i <- [0..n-1] ]

decode_interpretation :: Model -> Int -> String -> Int -> (String, (Int, [Int]))
decode_interpretation model k f n = (f, (a0, as))
  where 
    a0 = evalExp model (var_a0 k f)
    as = [ if evalFormula model (var_ai k f i) then 1 else 0 | i <- [0..n-1] ]

decode_algebra :: Model -> Int -> Signature -> Linear.Algebra
decode_algebra model k sig =
  [ decode_interpretation model k f n
  | (f, n) <- sig ]

decode :: Int -> Model -> Signature -> Order
decode k model sig =
  Linear (decode_algebra model k sig)

reduction_pair :: Int -> ReductionPair.ReductionPair
reduction_pair k = ReductionPair.ReductionPair {
  _side_condition = side_condition k,
  _geq = geq_A k,
  _gt = gt_A k,
  _decode = decode k,
  _monotone = f_i_monotone k,
  _invariant = f_i_invariant k
}
