module LinearGWPO where

-- GWPO with linear polynomial on N

import Control.Monad
import Data.List

import Term
import Rule
import TRS
import SMT
import Polynomial as P
import Status
import Order
import DP
import Usable

type Parameter = (P.Algebra, Status)

showParameter :: Parameter -> String
showParameter (a, st) =
  showAlgebra a ++
  showStatus st

geqGWPO :: Parameter -> Term -> Term -> Bool
geqGWPO (_, _) (V x) (V y) = x == y
geqGWPO (a, _) s@(V _) t@(F _ _) = P.gtA a (s, t)
geqGWPO param@(a, st) s@(F _ _) t@(V _) =
  P.gtA a (s, t) || (P.geqA a (s, t) && any (\s' -> geqGWPO param s' t) (args st s))
geqGWPO param@(a, st) s@(F _ _) t@(F _ _) = 
  P.gtA a (s, t) ||
    (P.geqA a (s, t) &&
      (any (\s' -> geqGWPO param s' t) (args st s) ||
       (all (\t' -> gtGWPO param s t') (args st t) &&
        (P.gtA a (bang s, bang t) || (P.geqA a (bang s, bang t) && geqGWPOLex param (args st s) (args st t))))))

gtGWPO :: Parameter -> Term -> Term -> Bool
gtGWPO (a, _) s@(V _) t = P.gtA a (s, t)
gtGWPO param@(a, st) s@(F _ _) t@(V _) =
  P.gtA a (s, t) || (P.geqA a (s, t) && any (\s' -> geqGWPO param s' t) (args st s))
gtGWPO param@(a, st) s@(F _ _) t@(F _ _) = 
  P.gtA a (s, t) ||
    (P.geqA a (s, t) &&
      (any (\s' -> geqGWPO param s' t) (args st s) ||
       (all (\t' -> gtGWPO param s t') (args st t) &&
        (P.gtA a (bang s, bang t) || (P.geqA a (bang s, bang t) && gtGWPOLex param (args st s) (args st t))))))

-- args are assumed to be already filtered
geqGWPOLex :: Parameter -> [Term] -> [Term] -> Bool
geqGWPOLex _ [] [] = True
geqGWPOLex _ [] (_ : _) = False
geqGWPOLex _ (_ : _) []  = True
geqGWPOLex param (s : ss) (t : ts) =
  gtGWPO param s t || 
  (geqGWPO param s t && geqGWPOLex param ss ts)

gtGWPOLex :: Parameter -> [Term] -> [Term] -> Bool
gtGWPOLex _ [] [] = False
gtGWPOLex _ [] (_ : _) = False
gtGWPOLex _ (_ : _) []  = True
gtGWPOLex param (s : ss) (t : ts) =
  gtGWPO param s t || 
  (geqGWPO param s t && gtGWPOLex param ss ts)

-- Encoding functions.

type Table = [(Term,Term)]

varF0 :: String -> Exp
varF0 f = Var ("a_0_" ++ f) 

varFi :: String -> Int -> Formula
varFi f i = FVar ("a_" ++ show i ++ "_" ++ f) 

varPrec :: String -> Exp
varPrec f = Var ("p_" ++ f)

varStatus :: String -> Int -> Formula
varStatus f i = FVar ("st_" ++ show i ++ "_" ++ f) 

varGeqGWPO :: Table -> Rule -> Formula
varGeqGWPO table rule
  | Just i <- elemIndex rule table = FVar ("geqGWPO_" ++ show i)
  | otherwise                      = error "varGeqGWPO"

varGtGWPO :: Table -> Rule -> Formula
varGtGWPO table rule
  | Just i <- elemIndex rule table = FVar ("gtGWPO_" ++ show i)
  | otherwise                      = error "varGtGWPO"

constantPart :: Term -> Exp
constantPart (V _)    = Val 0
constantPart (F f ts) =
  SMT.plus (varF0 f : [ times01 (varFi f i) (constantPart ti)
                      | (i,ti) <- zip [1..] ts ])

coefficient :: String -> Term -> Exp
coefficient x (V y)
  | x == y    = Val 1
  | otherwise = Val 0
coefficient x (F f ts) =
  SMT.plus [ times01 (varFi f i) (LinearGWPO.coefficient x ti) | (i, ti) <- zip [1..] ts ]

-- s >=_A t
encodeGeqA :: Rule -> Formula
encodeGeqA rule@(l, r) =
  conj (SMT.geq (constantPart l) (constantPart r) :
        [ SMT.geq (LinearGWPO.coefficient x l) (LinearGWPO.coefficient x r) 
        | x <- Rule.variables rule ])

-- s >_A t
encodeGtA :: Rule -> Formula
encodeGtA rule@(l, r) =
  conj (SMT.gt (constantPart l) (constantPart r) :
        [ SMT.geq (LinearGWPO.coefficient x l) (LinearGWPO.coefficient x r) 
        | x <- Rule.variables rule ])

-- s >_gwpo t
encodeGtGWPO :: Table -> Rule -> Formula
encodeGtGWPO _ rl@(V _, _) = encodeGtA rl
encodeGtGWPO table rl@(F _ _, V _) =
  disj [ encodeGtA rl, conj [ encodeGeqA rl, encodeGWPOSub table rl] ]
encodeGtGWPO table rl@(s@(F f ss), t@(F g ts)) =
  disj [ encodeGtA rl,
    conj [ encodeGeqA rl,
      disj [
        encodeGWPOSub table rl,
        conj [
          encodeGWPOArgs table rl,
          disj [
            encodeGtA (bang s, bang t),
            -- TODO: allow f ≠ g
            if f == g then conj [encodeGeqA (bang s, bang t), encodeGtGWPOLex table f ss ts] else bottom
          ]
        ]
      ]
    ]
  ]

-- s >=_gwpo t
encodeGeqGWPO :: Table -> Rule -> Formula
encodeGeqGWPO _ (V x, V y) = if x == y then top else bottom
encodeGeqGWPO _ rl@(V _, _) = encodeGtA rl
encodeGeqGWPO table rl@(F _ _, V _) =
  disj [ encodeGtA rl, conj [ encodeGeqA rl, encodeGWPOSub table rl] ]
encodeGeqGWPO table rl@(s@(F f ss), t@(F g ts)) =
  disj [ encodeGtA rl,
    conj [ encodeGeqA rl,
      disj [
        encodeGWPOSub table rl,
        conj [
          encodeGWPOArgs table rl,
          disj [
            encodeGtA (bang s, bang t),
            -- TODO: allow f ≠ g
            if f == g then conj [encodeGeqA (bang s, bang t), encodeGeqGWPOLex table f ss ts] else bottom
          ]
        ]
      ]
    ]
  ]

encodeGWPOSub :: Table -> Rule -> Formula
encodeGWPOSub table (F f ss, t) =
 disj [ conj [ varStatus f i, varGeqGWPO table (s', t) ] | (s', i) <- zip ss [1..] ]
encodeGWPOSub _ _ = error "encodeGWPOSub"

encodeGWPOArgs :: Table -> Rule -> Formula
encodeGWPOArgs table (s, F g ts) =
 conj [ implies (varStatus g j) (varGtGWPO table (s, t')) | (t', j) <- zip ts [1..] ]
encodeGWPOArgs _ _ = error "encodeGWPOArgs"

encodeGtGWPOLex :: Table -> String -> [Term] -> [Term] -> Formula
encodeGtGWPOLex tbl f ss ts = encodeGtGWPOLex' tbl f 1 ss ts

encodeGtGWPOLex' :: Table -> String -> Int -> [Term] -> [Term] -> Formula
encodeGtGWPOLex' _ _ _ [] [] = bottom
encodeGtGWPOLex' tbl f i (s : ss) (t : ts) =
  disj [ conj [ varStatus f i, varGtGWPO tbl (s, t) ],
         conj [ implies (varStatus f i) (varGeqGWPO tbl (s, t)), encodeGtGWPOLex' tbl f (i+1) ss ts ] ]
encodeGtGWPOLex' _ _ _ _ _ = error "length of terms must be the same"

encodeGeqGWPOLex :: Table -> String -> [Term] -> [Term] -> Formula
encodeGeqGWPOLex tbl f ss ts = encodeGeqGWPOLex' tbl f 1 ss ts

encodeGeqGWPOLex' :: Table -> String -> Int -> [Term] -> [Term] -> Formula
encodeGeqGWPOLex' _ _ _ [] [] = top
encodeGeqGWPOLex' tbl f i (s : ss) (t : ts) =
  disj [ conj [ varStatus f i, varGtGWPO tbl (s, t) ],
         conj [ implies (varStatus f i) (varGeqGWPO tbl (s, t)), encodeGeqGWPOLex' tbl f (i+1) ss ts ] ]
encodeGeqGWPOLex' _ _ _ _ _ = error "length of terms must be the same"

encode_one :: TRS -> TRS -> Formula
encode_one trs1 trs2 =
  conj (disj [ varGtGWPO table rl | rl <- trs1 ] :
        [ varGeqGWPO table rule | rule <- trs1 ++ trs2 ] ++
        [ SMT.geq (varF0 f) (Val 0) | (f, _) <- fs ] ++
        [ SMT.geq (varF0 (f ++ "!")) (Val 0) | (f, _) <- fs ] ++
         -- pi-simplicity
        [ implies (varStatus f i) (varFi f i) | (f, n) <- fs, i <- [1..n] ] ++
        [ implies (varGtGWPO table rule) (encodeGtGWPO table rule) | rule <- table ] ++
        [ implies (varGeqGWPO table rule) (encodeGeqGWPO table rule) | rule <- table ])
  where 
    fs = union (TRS.signatureOf trs1) (TRS.signatureOf trs2)
    table = nub [ (s', t') | (lhs, rhs) <- trs1 ++ trs2,
                             s' <- subterms lhs,
                             t' <- subterms rhs ]

encode_all :: TRS -> TRS -> Formula
encode_all trs1 trs2 =
  conj ([ varGtGWPO table rl | rl <- trs1 ] ++
        [ varGeqGWPO table rule | rule <- trs1 ++ trs2 ] ++
        [ SMT.geq (varF0 f) (Val 0) | (f, _) <- fs ] ++
        [ SMT.geq (varF0 (f ++ "!")) (Val 0) | (f, _) <- fs ] ++
         -- pi-simplicity
        [ implies (varStatus f i) (varFi f i) | (f, n) <- fs, i <- [1..n] ] ++
        [ implies (varGtGWPO table rule) (encodeGtGWPO table rule) | rule <- table ] ++
        [ implies (varGeqGWPO table rule) (encodeGeqGWPO table rule) | rule <- table ])
  where 
    fs = union (TRS.signatureOf trs1) (TRS.signatureOf trs2)
    table = nub [ (s', t') | (lhs, rhs) <- trs1 ++ trs2,
                             s' <- subterms lhs,
                             t' <- subterms rhs ]

-- Decoding functions

decodeInterpretation :: String -> Int -> Model -> Interpretation
decodeInterpretation f n model = (f, (f0, fs))
  where 
    f0 = evalExp model (varF0 f)
    fs = [ if evalFormula model (varFi f i) then 1 else 0 | i <- [1 .. n] ]

decodeAlgebra :: Signature -> Model -> Algebra
decodeAlgebra sig model =
  [ decodeInterpretation f n model | (f, n) <- sig ]

decodeStatus :: Signature -> Model -> Status
decodeStatus sig model =
  [ (f, [ evalFormula model (varStatus f i) | i <- [1..n] ]) | (f, n) <- sig ]

decode :: TRS -> TRS -> Model -> Parameter
decode r s model =
  (decodeAlgebra (sig ++ banged) model,
   decodeStatus sig model)
  where
    sig =  union (TRS.signatureOf r) (TRS.signatureOf s)
    banged = [ (f ++ "!", n) | (f, n) <- sig ]

solveOne :: String -> Problem -> IO (Maybe (Order, TRS, Problem))
solveOne tool (ps, rs, m)= do
  result <- sat tool (encode_one ps us)
  case result of
    Nothing -> return Nothing
    Just model 
      | length ps' < length ps ->
          return (Just (LinearGWPO a st, us, (ps', rs, m))) 
      | otherwise -> error "solveOne"
      where 
        params@(a, st) = decode ps us model
        ps' = [ rule | rule@(l, r) <- ps, not (gtGWPO params l r) ]
  where
    us = Usable.usable_rules ps rs

solveAll :: String -> Problem -> IO (Maybe Order)
solveAll tool (ps, rs, _)= do
  result <- sat tool (encode_all ps rs)
  case result of
    Nothing -> return Nothing
    Just model -> do
          forM_ ps checkGt
          forM_ rs checkGeq
          return (Just (LinearGWPO a st)) 
      where 
        param@(a, st) = decode ps rs model
        checkGt (s, t)
          | gtGWPO param s t = return ()
          | otherwise = error (show s ++ " > " ++ show t ++ " is not satisfied")
        checkGeq (s, t)
          | geqGWPO param s t = return ()
          | otherwise = error (show s ++ " >= " ++ show t ++ " is not satisfied")
