module Main where

import Data.List.Split (splitOn)
import TRS
import DP
import Proof
import qualified Recursive as R
import qualified Matrix
import ReductionPair
import qualified Lex
import qualified EncodingMaxPlus
import qualified EncodingMatrix
import ARIParser
import qualified LPO
import Options.Applicative

data Opt = Opt { 
  _orders :: String,
  _file :: String,
  _no_rule_removal :: Bool,
  _no_usable_rules :: Bool,
  _oneshot :: Bool
}

option_spec :: Parser Opt
option_spec = Opt
  <$> strOption 
    (long "orders"
     <> short 'o'
     <> metavar "ORDERS"
     <> value "l"
     <> help ("orderings: l, m, ms[2345], ml[2345], lpo (e.g, -o ml4+lpo,m)"))
  <*> strOption
    (long "file"
     <> short 'f'
     <> metavar "TARGET"
     <> help "target file in ARI/WST format")
  <*> switch
    (long "no-rule-removal"
     <> help "do not use the rule removal criterion")
  <*> switch
    (long "no-usable-rules"
     <> help "do not use the usable rule criterion")
  <*> switch
    (long "oneshot"
     <> help "strictly orient all rules in DP(R)")

-- TODO: nicely print this in help
orderDesc :: String
orderDesc =
  "  lpo  LPO with argument filtering" ++
  "  l    linear interpretations on N equipped" ++
  "         f_A(x_1,...,x_n) = f_0 + f_1 x_1 + ... + f_n x_n" ++
  "         where f_0 in N, f_1,...,f_n in {0,1}" ++
  "  m    max/plus interpretations on N:" ++
  "         f_A(x_1, ..., x_n) = max{f_0, f_1 x_1 + c_1, ..., f_n x_n + c_n}" ++
  "         where f_0 in N, f_1,...,f_n in {0,1}, and c_1,...,c_n in Z" ++
  "  wpol   WPO with linear polynomials" ++
  "  gwpol  GWPO with linear polynomials" ++
  "  spol   (M)SPO with linear polynomials" ++
  "  wpom   WPO with max/plus" ++
  "  gwpom  GWPO with max/plus" ++
  "  spom   (M)SPO with max/plus" ++
  "  l1lpo  lexicographic combination of l1 and LPO (as single reduction pair)" ++
  "  mlpo   lexicographic combination of max/plus and LPO (as single reduction pair)"

print_proof :: (TRS, TRS, Proof) -> IO ()
print_proof (rs, [], proof) = do
  putStrLn "We show the termination of the TRS R:"
  putStrLn ""
  putStr (showTRS rs)
  putStrLn ""
  putStr (show_proof proof)
print_proof (rs, ss, proof) = do
  putStrLn "We show the termination of the relative TRS R/S:"
  putStrLn ""
  putStrLn "  R:"
  putStr (showTRS rs)
  putStrLn ""
  putStrLn "  S:"
  putStr (showTRS ss)
  putStrLn ""
  putStr (show_proof proof)

parse_base_order :: Int -> String -> ReductionPair
parse_base_order k "l"   = EncodingMatrix.reduction_pair Matrix.Standard 1 k
parse_base_order k "ms2" = EncodingMatrix.reduction_pair Matrix.Standard 2 k
parse_base_order k "ms3" = EncodingMatrix.reduction_pair Matrix.Standard 3 k
parse_base_order k "ms4" = EncodingMatrix.reduction_pair Matrix.Standard 4 k
parse_base_order k "ms5" = EncodingMatrix.reduction_pair Matrix.Standard 5 k
parse_base_order k "ml2" = EncodingMatrix.reduction_pair Matrix.Lex 2 k
parse_base_order k "ml3" = EncodingMatrix.reduction_pair Matrix.Lex 3 k
parse_base_order k "ml4" = EncodingMatrix.reduction_pair Matrix.Lex 4 k
parse_base_order k "ml5" = EncodingMatrix.reduction_pair Matrix.Lex 5 k
parse_base_order k "m"   = EncodingMaxPlus.reduction_pair k
parse_base_order k "lpo" = LPO.reduction_pair k
parse_base_order _ _     = error "parse_base_order"

parse_order :: String -> ReductionPair
parse_order s =
  case splitOn "+" s of
    []  -> error "Specify base orders."
    [name] -> parse_base_order 0 name
    names  -> Lex.lex [ parse_base_order i name 
                      | (i, name) <- zip [0..] names ]

parse_orders :: String -> [ReductionPair]
parse_orders s =
  case splitOn "," s of
    []    -> error "Specify base orders."
    names -> [ parse_order name | name <- names ]

prove_recursive :: Bool -> [R.Criterion] -> TRS -> TRS -> IO ()
prove_recursive oneshot criteria rs ss =
  case dp_problem rs ss of
    Nothing -> putStrLn "MAYBE"
    Just problem -> do
      result <- R.solve_problem oneshot criteria problem 
      case result of
        Nothing    -> do
          putStrLn "MAYBE"
        Just proof -> do
          putStrLn "YES"
          putStrLn ""
          print_proof (rs, ss, proof)

prove :: ReductionPair.Option -> [ReductionPair] -> TRS -> TRS -> IO ()
prove rp_option rps rs ss =
  prove_recursive 
    (_all_strict rp_option)
    [ process_with rp_option rp "z3" | rp <- rps] rs ss

dispatch :: Opt -> IO ()
dispatch opt = do
  errorOrTRS <- read_file (_file opt)
  case errorOrTRS of
    Left e    -> putStrLn (show e)
    Right (rs, ss)
      | TRS.variableCondition (rs ++ ss) -> 
        let rp_option = Option{
          _all_strict       = _oneshot opt,
          _use_rule_removal = not (_no_rule_removal opt),
          _use_usable_rules = not (_no_usable_rules opt)
          }
        in 
          prove rp_option (parse_orders (_orders opt)) rs ss
      | otherwise -> do
          putStrLn "NO"
          putStrLn ""
          putStrLn "The input TRS violates the variable condition and therefore is non-terminating."

main :: IO ()
main = dispatch =<< execParser opts
  where
    opts = info (option_spec <**> helper)
      (fullDesc <> header "Mint -- Mini Termination Tool verion 0.3" )
