-- Matrix interpretations equipped with
--  - the standard order on vectors
--  - the lexicographic order on vectors 
module Matrix where

import Data.List

type Vector = [Int]
type Matrix = [Vector]
type Linear = (Int, [Int])
type Form = (Vector, [Matrix])
type Interpretation = (String, Form)
data BaseOrder = Standard | Lex
type Algebra = (BaseOrder, Int, [Interpretation])

is_zero_vector :: Vector -> Bool
is_zero_vector v = all (== 0) v

is_zero_matrix :: Matrix -> Bool
is_zero_matrix m = all is_zero_vector m

is_unit_matrix :: Matrix -> Bool
is_unit_matrix m =
  and [ (i == j && m_ij == 1) || m_ij == 0
      | i <- [0..d-1],
        j <- [0..d-1],
        let m_ij = m !! i !! j ]
  where d = length m

show_vector :: Vector -> String
show_vector [a] = show a
show_vector v   = 
  "(" ++ intercalate "," [ show a | a <- v ] ++ ")"

show_matrix :: Matrix -> String
show_matrix [[a]] = show a
show_matrix m =
  "(" ++ intercalate "," [ show_vector v | v <- m ] ++ ")" 

show_m_xi :: (Matrix, Int) -> String
show_m_xi (m, i)
  | is_unit_matrix m = "x" ++ show i
  | otherwise        = show_matrix m ++ " x" ++ show i

show_form :: Form -> String
show_form (v, ms)
  | ss == [] = s
  | is_zero_vector v = intercalate " + " ss
  | otherwise        = intercalate " + " (ss ++ [s])
  where
    s  = show_vector v
    ss = [ show_m_xi (m, i) 
         | (m, i) <- zip ms [1 :: Int ..],
           not (is_zero_matrix m) ]

show_lhs :: (String, Int) -> String
show_lhs (f, n) =
  f ++ "_A(" ++ intercalate "," [ "x" ++ show i | i <- [1..n] ] ++ ")"

show_interpretation :: Interpretation -> String
show_interpretation (f, form@(_, ms)) =
  show_lhs (f, length ms) ++ " = " ++ show_form form

show_order :: BaseOrder -> String
show_order Standard = "standard order"
show_order Lex = "lexicographic order"

show_algebra :: Algebra -> String
show_algebra (order, d, interpretations) =
  unlines 
    (["  carrier: N^" ++ show d,
      "  order: " ++ show_order order,
      "  interpretations:" ] ++
     [ "    " ++ show_interpretation f_A | f_A <- interpretations ])


{-
add_vector :: Vector -> Vector -> Vector
add_vector u v = zipWith (+) u v

sum_vector :: Int -> [Vector] -> Vector
sum_vector d vs = foldl add_vector (zero_vector d) vs

zero_matrix :: Int -> Matrix
zero_matrix d = replicate d (zero_vector d)

unit_matrix :: Int -> Matrix
unit_matrix d =
  [ [ if i == j then 1 else 0 | j <- [0..d-1] ] 
  | i <- [0..d-1] ]

add_matrix :: Matrix -> Matrix -> Matrix
add_matrix m1 m2 = zipWith add_vector m1 m2

sum_matrix :: Int -> [Matrix] -> Matrix
sum_matrix d ms = foldl add_matrix (zero_matrix d) ms

inner_product :: Vector -> Vector -> Int
inner_product v1 v2 =
  sum (0 : zipWith (*) v1 v2)

mul_matrix_vector :: Matrix -> Vector -> Vector
mul_matrix_vector m v = [ inner_product u v | u <- m ]

mul_vector_matrix :: Vector -> Matrix -> Vector
mul_vector_matrix u m = [ inner_product u v | v <- m ]

mul_matrix :: Matrix -> Matrix -> Matrix
mul_matrix m1 m2 =
  [ mul_vector_matrix v1 (transpose m2) | v1 <- m1 ]

-- C_A(t,x) returns the coefficient matrix of alpha(x) in [\alpha]_A(t):
-- C_A(x,x) = E
-- C_A(y,x) = O
-- C_A(f(t1,...,tn),x) = F1 C_A(t1,x) + ... + Fn C_A(tn,x),
--   where f_A(x1,...,xn) = f0 + F1 x1 + ... + Fn xn
coefficient :: Algebra -> Term -> String -> Matrix
coefficient (_, d, _) (V x) y
  | x == y    = unit_matrix d
  | otherwise = zero_matrix d
coefficient a@(_, d, interpretations) (F f ts) x =
  case lookup f interpretations of
    Nothing -> error "coefficient"
    Just (_, _Fs) -> 
      sum_matrix d [ mul_matrix _Fk (coefficient a tk x) 
                 | (_Fk, tk) <- zip _Fs ts ]

-- c_A(t) returns the constant part of [\alpha]_A(t):
-- c_A(x) = 0
-- c_A(f(t1,...,tn)) = f0 + c_A(t1) + ... + c_A(tn)
--   where f_A(x1,...,xn) = f0 + F1 x1 + ... + Fn xn
constant :: Algebra -> Term -> Vector
constant (_, d, _) (V _) = zero_vector d
constant a@(_, d, interpretations) (F f ts) =
  case lookup f interpretations of
    Nothing -> error "constant"
    Just (f0, _Fs) -> sum_vector d (f0 : vs)
      where
        vs = [ mul_matrix_vector _Fk (constant a tk) 
             | (_Fk, tk) <- zip _Fs ts ]

geq_linear :: Linear -> Linear -> Bool 
geq_linear (a0, as) (b0, bs) = 
  a0 >= b0 && all geq (zip as bs) 
  where geq (ak,bk) = ak >= bk

gt_linear :: Linear -> Linear -> Bool 
gt_linear (a0, as) (b0, bs) = 
  a0 > b0 && all geq (zip as bs) 
  where geq (ak,bk) = ak >= bk

geq_linear_lex :: [Linear] -> [Linear] -> Bool 
geq_linear_lex []         []         = True
geq_linear_lex (l1 : ls1) (l2 : ls2) =
  gt_linear l1 l2 || 
  (geq_linear l1 l2 && geq_linear_lex ls1 ls2)
geq_linear_lex _ _ = error "geq_linear_lex"

gt_linear_lex :: [Linear] -> [Linear] -> Bool 
gt_linear_lex []         []         = False
gt_linear_lex (l1 : ls1) (l2 : ls2) =
  gt_linear l1 l2 || 
  (geq_linear l1 l2 && gt_linear_lex ls1 ls2)
gt_linear_lex _ _ = error "gt_linear_lex"

geq_linears :: BaseOrder -> [Linear] -> [Linear] -> Bool
geq_linears Lex      ls1 ls2 = geq_linear_lex ls1 ls2
geq_linears Standard ls1 ls2 = 
  all (\(l1, l2) -> geq_linear l1 l2) (zip ls1 ls2)

gt_linears :: BaseOrder -> [Linear] -> [Linear] -> Bool
gt_linears Lex ls1 ls2 = gt_linear_lex ls1 ls2
gt_linears Standard (l1 : ls1) (l2 : ls2) = 
  gt_linear l1 l2 &&
  all (\(l1', l2') -> geq_linear l1' l2') (zip ls1 ls2)
gt_linears Standard _ _ = error "gt_linears"

interpret :: Algebra -> [String] -> Term -> [Linear]
interpret a xs t = 
  [ (v !! i, [ m_ij | m <- ms, m_ij <- m !! i ] )
  | i <- [0 .. length v - 1] ]
  where 
    v = constant a t
    ms = [ coefficient a t x | x <- xs ]

-- >=_A
geq_A :: Algebra -> Rule -> Bool
geq_A a@(order, _, _) (s, t) =
  geq_linears order (interpret a xs s) (interpret a xs t)
  where xs = Rule.variables (s, t)

-- >=_A
gt_A :: Algebra -> Rule -> Bool
gt_A a@(order, _, _) (s, t) =
  gt_linears order (interpret a xs s) (interpret a xs t)
  where xs = Rule.variables (s, t)

-- >>> f_A = ("f", ([0,1], [[[1,1],[1,0]]]))
-- >>> g_A = ("g", ([0,0], [[[1,0],[0,0]]]))
-- >>> a = (Lex,2,[f_A,g_A]) 
-- >>> gt_A a (read_term "f(f(x))", read_term "f(g(f(x)))")
-- True
--
-}