module MaxPlusLPO where

-- lex combination of max/plus and LPO

import Control.Monad
import Data.List

import Term
import Rule
import TRS
import DP
import SMT
import MaxPlus as MP
import SymbolicMaxPlus as SMaxPlus
import Usable
import Order
import Precedence


type Parameter = (MP.Algebra, Precedence)

-- combination of >_A and LPO
-- note that max/plus interpretations are never strictly monotone
-- at any positions
-- therefore, empty status is the only possible status;
-- in other words, possible LPO is simply comparison of head symbols by precedence

geqALPO :: Parameter -> Term -> Term -> Bool
geqALPO (a, p) s@(F _ _) t@(F _ _) =
  MP.gtA a (s, t) || (MP.geqA a (s, t) && geqPrec p s t)
geqALPO (a, _) s t = MP.gtA a (s, t)

gtALPO :: Parameter -> Term -> Term -> Bool
gtALPO (a, p) s@(F _ _) t@(F _ _) =
  MP.gtA a (s, t) || (MP.geqA a (s, t) && gtPrec p s t)
gtALPO (a, _) s t = MP.gtA a (s, t)

-- Encoding functions.

varF0 :: String -> Exp
varF0 f = Var ("a0_" ++ f) 

varFi :: String -> Int -> Exp
varFi f i = Var ("a_" ++ show i ++ "_" ++ f) 

varF'i :: String -> Int -> Formula
varF'i f i = FVar ("b_" ++ show i ++ "_" ++ f) 

symbolicMaxPlus :: Term -> SymbolicMaxPlus
symbolicMaxPlus (V x) = SMaxPlus.var x
symbolicMaxPlus (F f ts) =
   SMaxPlus.flatten (SMaxPlus.const (varF0 f) : [ multiplyAdd f i ti | (i, ti) <- zip [1..] ts ])
  where
    multiplyAdd fname i ti = SMaxPlus.addConstant (varFi fname i) (SMaxPlus.mult (varF'i fname i) (symbolicMaxPlus ti))

-- s >=_A t
encodeGeqA :: Rule -> Formula
encodeGeqA (l, r) = SMaxPlus.encodeGeq (symbolicMaxPlus l) (symbolicMaxPlus r)

-- s >_A t
encodeGtA :: Rule -> Formula
encodeGtA (l, r) = SMaxPlus.encodeGt (symbolicMaxPlus l) (symbolicMaxPlus r)

varPrec :: String -> Exp
varPrec f = Var ("p_" ++ f)

-- f >= g
encodeGeqPrec :: Rule -> Formula
encodeGeqPrec (F f _, F g _) = SMT.geq (varPrec f) (varPrec g)
encodeGeqPrec _ = error "encodeGeqPrec"

-- f > g
encodeGtPrec :: Rule -> Formula
encodeGtPrec (F f _, F g _) = SMT.gt (varPrec f) (varPrec g)
encodeGtPrec _ = error "encodeGtPrec"

encodeGeqALPO :: Rule -> Formula
encodeGeqALPO rule@(F _ _, F _ _) =
  disj [encodeGtA rule, conj [encodeGeqA rule, encodeGeqPrec rule]]
encodeGeqALPO rule = encodeGtA rule

encodeGtALPO :: Rule -> Formula
encodeGtALPO rule@(F _ _, F _ _) =
  disj [encodeGtA rule, conj [encodeGeqA rule, encodeGtPrec rule]]
encodeGtALPO rule = encodeGtA rule

encode_one :: TRS -> TRS -> Formula
encode_one trs1 trs2 = 
  conj ([ SMT.geq (varF0 f) (Val 0) | (f, _) <- sig ] ++
        [ encodeGeqALPO rule | rule <- trs1 ++ trs2 ] ++
        [ disj [ encodeGtALPO rule | rule <- trs1 ] ])
  where
    sig = TRS.signatureOf (trs1 ++ trs2)

encode_all :: TRS -> TRS -> Formula
encode_all trs1 trs2 = 
  conj ([ SMT.geq (varF0 f) (Val 0) | (f, _) <- sig ] ++
        [ encodeGeqALPO rule | rule <- trs1 ++ trs2 ] ++
        [ encodeGtALPO rule | rule <- trs1 ])
  where
    sig = TRS.signatureOf (trs1 ++ trs2)

-- decoding

decodeInterpretation :: String -> Int -> Model -> Interpretation
decodeInterpretation f n model = (f, (f0, ff's))
  where 
    f0 = evalExp model (varF0 f)
    evalf fname i = evalExp model (varFi fname i)
    evalf' fname i = if evalFormula model (varF'i fname i) then 1 else 0
    ff's = [ (evalf f i, evalf' f i) | i <- [1 .. n] ]

decodeAlgebra :: Signature -> Model -> Algebra
decodeAlgebra sig model =
  [ decodeInterpretation f n model | (f, n) <- sig ]

decodePrecedence :: Signature -> Model -> Precedence
decodePrecedence sig model =
  [ (f, evalExp model (varPrec f)) | (f, _) <- sig ]

decode :: TRS -> TRS -> Model -> Parameter
decode r s model =
  (decodeAlgebra sig model,
   decodePrecedence sig model)
  where
    sig =  union (TRS.signatureOf r) (TRS.signatureOf s)

solveOne :: String -> Problem -> IO (Maybe (Order, TRS, Problem))
solveOne tool (ps, rs, m) = do
  result <- sat tool (encode_one ps us)
  case result of
    Nothing           -> return Nothing
    Just model 
      | length ps == length ps' -> error "solveOne"
      | otherwise               -> return (Just (MaxPlusLPO a p, us, (ps', rs, m)))
      where 
        (a, p) = decode ps us model
        ps' = [ rule | rule@(s, t) <- ps, not (gtALPO (a, p) s t) ]
  where 
    us = usable_rules ps rs

solveAll :: String -> Problem -> IO (Maybe Order)
solveAll tool (ps, rs, _) = do
  result <- sat tool (encode_all ps rs)
  case result of
    Nothing -> return Nothing
    Just model -> do
          forM_ ps checkGt
          forM_ rs checkGeq
          return (Just (MaxPlusLPO a p)) 
      where 
        param@(a, p) = decode ps rs model
        checkGt (s, t)
          | gtALPO param s t = return ()
          | otherwise = error (show s ++ " > " ++ show t ++ " is not satisfied")
        checkGeq (s, t)
          | geqALPO param s t = return ()
          | otherwise = error (show s ++ " >= " ++ show t ++ " is not satisfied")
