module MaxPlusSPO where

-- SPO with linear polynomial on N
-- strict precedence (not quasi-)

import Control.Monad
import Data.List

import Term
import Rule
import TRS
import SMT
import MaxPlus as MP
import SymbolicMaxPlus as SMaxPlus
import Precedence
import Status
import Order
import DP
import Usable

type Parameter = (MP.Algebra, Precedence, Status)

showParameter :: Parameter -> String
showParameter (a, p, st) =
  showAlgebra a ++
  showPrecedence p ++
  showStatus st

-- underlying orders (\sqsupseteq, \sqsupset) resembling WPO
geqSq :: Parameter -> Term -> Term -> Bool
geqSq (a, pr, _) s@(F _ _) t@(F _ _) =
  MP.gtA a (bang s, bang t) || (MP.geqA a (bang s, bang t) && geqPrec pr s t)
geqSq _ s t = s == t

gtSq :: Parameter -> Term -> Term -> Bool
gtSq (a, pr, _) s@(F _ _) t@(F _ _) =
  MP.gtA a (bang s, bang t) || (geqA a (bang s, bang t) && gtPrec pr s t)
gtSq _ _ _ = False

geqSPO :: Parameter -> Term -> Term -> Bool
geqSPO _ (V x) (V y) = x == y
geqSPO _ (V _) (F _ _) = False
geqSPO param@(_, _, st) s@(F _ _) t@(V _) =
  any (\s' -> geqSPO param s' t) (args st s)
geqSPO param@(_, _, st) s@(F _ _) t@(F _ _) = 
  any (\s' -> geqSPO param s' t) (args st s) ||
  (all (\t' -> gtSPO param s t') (args st t) &&
    (gtSq param s t || (geqSq param s t && geqSPOLex param (args st s) (args st t))))

gtSPO :: Parameter -> Term -> Term -> Bool
gtSPO _ (V _) _ = False
gtSPO param@(_, _, st) s@(F _ _) t@(V _) =
  any (\s' -> geqSPO param s' t) (args st s)
gtSPO param@(_, _, st) s@(F _ _) t@(F _ _) = 
  any (\s' -> geqSPO param s' t) (args st s) ||
  (all (\t' -> gtSPO param s t') (args st t) &&
    (gtSq param s t || (geqSq param s t && gtSPOLex param (args st s) (args st t))))

-- args are assumed to be already filtered
geqSPOLex :: Parameter -> [Term] -> [Term] -> Bool
geqSPOLex _ [] [] = True
geqSPOLex _ [] (_ : _) = False
geqSPOLex _ (_ : _) []  = True
geqSPOLex param (s : ss) (t : ts) =
  gtSPO param s t || 
  (geqSPO param s t && geqSPOLex param ss ts)

gtSPOLex :: Parameter -> [Term] -> [Term] -> Bool
gtSPOLex _ [] [] = False
gtSPOLex _ [] (_ : _) = False
gtSPOLex _ (_ : _) []  = True
gtSPOLex param (s : ss) (t : ts) =
  gtSPO param s t || 
  (geqSPO param s t && gtSPOLex param ss ts)

geqMSPO :: Parameter -> Term -> Term -> Bool
geqMSPO param@(a, _, _) s t = MP.geqA a (s, t) && geqSPO param s t

gtMSPO :: Parameter -> Term -> Term -> Bool
gtMSPO param@(a, _, _) s t = MP.geqA a (s, t) && gtSPO param s t

-- Encoding functions.


type Table = [(Term,Term)]

-- example:
-- f_A(x_1, x_2) = max { a_0_f, b_1_f x_1 + a_1_f, b_2_f x_2 + a_2_f }

varF0 :: String -> Exp
varF0 f = Var ("a_0_" ++ f) 

varFi :: String -> Int -> Exp
varFi f i = Var ("a_" ++ show i ++ "_" ++ f) 

varF'i :: String -> Int -> Formula
varF'i f i = FVar ("b_" ++ show i ++ "_" ++ f) 

varPrec :: String -> Exp
varPrec f = Var ("p_" ++ f)

varStatus :: String -> Int -> Formula
varStatus f i = FVar ("st_" ++ show i ++ "_" ++ f) 

varGeqSPO :: Table -> Rule -> Formula
varGeqSPO table rule
  | Just i <- elemIndex rule table = FVar ("geqSPO_" ++ show i)
  | otherwise                      = error "varGeqSPO"

varGtSPO :: Table -> Rule -> Formula
varGtSPO table rule
  | Just i <- elemIndex rule table = FVar ("gtSPO_" ++ show i)
  | otherwise                      = error "varGtSPO"

symbolicMaxPlus :: Term -> SymbolicMaxPlus
symbolicMaxPlus (V x) = SMaxPlus.var x
symbolicMaxPlus (F f ts) =
   SMaxPlus.flatten (SMaxPlus.const (varF0 f) : [ multiplyAdd f i ti | (i, ti) <- zip [1..] ts ])
  where
    multiplyAdd fname i ti = SMaxPlus.addConstant (varFi fname i) (SMaxPlus.mult (varF'i fname i) (symbolicMaxPlus ti))

-- s >=_A t
encodeGeqA :: Rule -> Formula
encodeGeqA (l, r) = SMaxPlus.encodeGeq (symbolicMaxPlus l) (symbolicMaxPlus r)

-- s >_A t
encodeGtA :: Rule -> Formula
encodeGtA (l, r) = SMaxPlus.encodeGt (symbolicMaxPlus l) (symbolicMaxPlus r)

-- f >= g
encodeGeqPrec :: Rule -> Formula
encodeGeqPrec (F f _, F g _) = SMT.geq (varPrec f) (varPrec g)
encodeGeqPrec _ = error "encodeGeqPrec"

-- f > g
encodeGtPrec :: Rule -> Formula
encodeGtPrec (F f _, F g _) = SMT.gt (varPrec f) (varPrec g)
encodeGtPrec _ = error "encodeGtPrec"

encodeGeqSq :: Rule -> Formula
encodeGeqSq (s@(F _ _), t@(F _ _)) =
  disj [ encodeGtA (bang s, bang t), conj [ encodeGeqA (bang s, bang t), encodeGeqPrec (s, t) ] ]
encodeGeqSq (s, t)
  | s == t = top
  | otherwise = bottom

encodeGtSq :: Rule -> Formula
encodeGtSq (s@(F _ _), t@(F _ _)) =
  disj [ encodeGtA (bang s, bang t), conj [ encodeGeqA (bang s, bang t), encodeGtPrec (s, t) ] ]
encodeGtSq _ = bottom

-- s >=_spo t
encodeGeqSPO :: Table -> Rule -> Formula
encodeGeqSPO _ (s@(V _), t)
  | s == t = top
  | otherwise = bottom
encodeGeqSPO table rl@(F _ _, V _) = encodeSPOSub table rl
encodeGeqSPO table rl@(F f ss, F g ts) =
  disj [
    encodeSPOSub table rl,
    conj [
      encodeSPOArgs table rl,
      disj [
        encodeGtSq rl,
        -- only strict precedence is considered for now
        if f == g then conj [encodeGeqSq rl, encodeGeqSPOLex table f ss ts] else bottom
      ]
    ]
  ]

-- s >_spo t
encodeGtSPO :: Table -> Rule -> Formula
encodeGtSPO _ (V _, _) = bottom
encodeGtSPO table rl@(F _ _, V _) = encodeSPOSub table rl
encodeGtSPO table rl@(F f ss, F g ts) =
  disj [
    encodeSPOSub table rl,
    conj [
      encodeSPOArgs table rl,
      disj [
        encodeGtSq rl,
        -- only strict precedence is considered for now
        if f == g then conj [encodeGeqSq rl, encodeGtSPOLex table f ss ts] else bottom
      ]
    ]
  ]

encodeSPOSub :: Table -> Rule -> Formula
encodeSPOSub table (F f ss, t) =
 disj [ conj [ varStatus f i, varGeqSPO table (s', t) ] | (s', i) <- zip ss [1..] ]
encodeSPOSub _ _ = error "encodeSPOSub"

encodeSPOArgs :: Table -> Rule -> Formula
encodeSPOArgs table (s, F g ts) =
 conj [ implies (varStatus g j) (varGtSPO table (s, t')) | (t', j) <- zip ts [1..] ]
encodeSPOArgs _ _ = error "encodeSPOArgs"

encodeGtSPOLex :: Table -> String -> [Term] -> [Term] -> Formula
encodeGtSPOLex tbl f ss ts = encodeGtSPOLex' tbl f 1 ss ts

encodeGtSPOLex' :: Table -> String -> Int -> [Term] -> [Term] -> Formula
encodeGtSPOLex' _ _ _ [] [] = bottom
encodeGtSPOLex' tbl f i (s : ss) (t : ts) =
  disj [ conj [ varStatus f i, varGtSPO tbl (s, t) ],
         conj [ implies (varStatus f i) (varGeqSPO tbl (s, t)), encodeGtSPOLex' tbl f (i+1) ss ts ] ]
encodeGtSPOLex' _ _ _ _ _ = error "length of terms must be the same"

encodeGeqSPOLex :: Table -> String -> [Term] -> [Term] -> Formula
encodeGeqSPOLex tbl f ss ts = encodeGeqSPOLex' tbl f 1 ss ts

encodeGeqSPOLex' :: Table -> String -> Int -> [Term] -> [Term] -> Formula
encodeGeqSPOLex' _ _ _ [] [] = top
encodeGeqSPOLex' tbl f i (s : ss) (t : ts) =
  disj [ conj [ varStatus f i, varGtSPO tbl (s, t) ],
         conj [ implies (varStatus f i) (varGeqSPO tbl (s, t)), encodeGeqSPOLex' tbl f (i+1) ss ts ] ]
encodeGeqSPOLex' _ _ _ _ _ = error "length of terms must be the same"

encodeGeqMSPO :: Table -> Rule -> Formula
encodeGeqMSPO tbl rl = conj [encodeGeqA rl, encodeGeqSPO tbl rl ]

encodeGtMSPO :: Table -> Rule -> Formula
encodeGtMSPO tbl rl = conj [encodeGeqA rl, encodeGtSPO tbl rl ]

encode_one :: TRS -> TRS -> Formula
encode_one trs1 trs2 =
  conj (disj [ varGtSPO table rl | rl <- trs1 ] :
        [ encodeGeqMSPO table rule | rule <- trs1 ++ trs2 ] ++
        [ SMT.geq (varF0 f) (Val 0) | (f, _) <- fs ] ++
        [ SMT.geq (varF0 (f ++ "!")) (Val 0) | (f, _) <- fs ] ++
        -- strict precedence 
        [ Distinct [ varPrec f | (f, _) <- fs ] ] ++
        [ implies (varGtSPO table rule) (encodeGtSPO table rule) | rule <- table ] ++
        [ implies (varGeqSPO table rule) (encodeGeqSPO table rule) | rule <- table ])
  where 
    fs = union (TRS.signatureOf trs1) (TRS.signatureOf trs2)
    table = nub [ (s', t') | (lhs, rhs) <- trs1 ++ trs2,
                             s' <- subterms lhs,
                             t' <- subterms rhs ]

encode_all :: TRS -> TRS -> Formula
encode_all trs1 trs2 =
  conj ([ varGtSPO table rl | rl <- trs1 ] ++
        [ encodeGeqMSPO table rule | rule <- trs1 ++ trs2 ] ++
        [ SMT.geq (varF0 f) (Val 0) | (f, _) <- fs ] ++
        [ SMT.geq (varF0 (f ++ "!")) (Val 0) | (f, _) <- fs ] ++
        -- strict precedence 
        [ Distinct [ varPrec f | (f, _) <- fs ] ] ++
        [ implies (varGtSPO table rule) (encodeGtSPO table rule) | rule <- table ] ++
        [ implies (varGeqSPO table rule) (encodeGeqSPO table rule) | rule <- table ])
  where 
    fs = union (TRS.signatureOf trs1) (TRS.signatureOf trs2)
    table = nub [ (s', t') | (lhs, rhs) <- trs1 ++ trs2,
                             s' <- subterms lhs,
                             t' <- subterms rhs ]


-- Decoding functions

decodeInterpretation :: String -> Int -> Model -> MP.Interpretation
decodeInterpretation f n model = (f, (f0, ff's))
  where 
    f0 = evalExp model (varF0 f)
    evalf fname i = evalExp model (varFi fname i)
    evalf' fname i = if evalFormula model (varF'i fname i) then 1 else 0
    ff's = [ (evalf f i, evalf' f i) | i <- [1 .. n] ]

decodeAlgebra :: Signature -> Model -> Algebra
decodeAlgebra sig model =
  [ decodeInterpretation f n model | (f, n) <- sig ]

decodePrecedence :: Signature -> Model -> Precedence
decodePrecedence sig model =
  [ (f, evalExp model (varPrec f)) | (f, _) <- sig ]

decodeStatus :: Signature -> Model -> Status
decodeStatus sig model =
  [ (f, [ evalFormula model (varStatus f i) | i <- [1..n] ]) | (f, n) <- sig ]

decode :: TRS -> TRS -> Model -> Parameter
decode trs1 trs2 model =
  (decodeAlgebra (sig ++ banged) model,
   decodePrecedence sig model,
   decodeStatus sig model)
  where
    sig =  union (TRS.signatureOf trs1) (TRS.signatureOf trs2)
    banged = [ (f ++ "!", n) | (f, n) <- sig ]

solveOne :: String -> Problem -> IO (Maybe (Order, TRS, Problem))
solveOne tool (ps, rs, m) = do
  result <- sat tool (encode_one ps us)
  case result of
    Nothing -> return Nothing
    Just model 
      | length ps' < length ps ->
          return (Just (MaxPlusSPO a p st, us, (ps', rs, m))) 
      | otherwise -> error "solveOne"
      where 
        params@(a, p, st) = decode ps us model
        ps' = [ rule | rule@(l, r) <- ps, not (gtSPO params l r) ]
  where
    us = Usable.usable_rules ps rs

solveAll :: String -> Problem -> IO (Maybe Order)
solveAll tool (ps, rs, _) = do
  result <- sat tool (encode_all ps rs)
  case result of
    Nothing -> return Nothing
    Just model -> do
          forM_ ps checkGt
          forM_ rs checkGeq
          return (Just (MaxPlusSPO a p st)) 
      where 
        param@(a, p, st) = decode ps rs model
        checkGt (s, t)
          | gtSPO param s t = return ()
          | otherwise = error (show s ++ " > " ++ show t ++ " is not satisfied")
        checkGeq (s, t)
          | geqMSPO param s t = return ()
          | otherwise = error (show s ++ " >= " ++ show t ++ " is not satisfied")
