module MaxPlusWPO where

-- WPO with max/plus on N
-- strict precedence (not quasi-)

import Control.Monad
import Data.List

import Term
import TRS
import Rule
import SMT
import MaxPlus as MP
import SymbolicMaxPlus as SMaxPlus
import Precedence
import Status
import Order
import DP
import Usable

type Parameter = (MP.Algebra, Precedence, Status)

showParameter :: Parameter -> String
showParameter (a, p, st) =
  MP.showAlgebra a ++
  showPrecedence p ++
  showStatus st

geqWPO :: Parameter -> Term -> Term -> Bool
geqWPO (_, _, _) (V x) (V y) = x == y
geqWPO (a, _, _) s@(V _) t@(F _ _) = MP.gtA a (s, t)
geqWPO param@(a, _, st) s@(F _ _) t@(V _) =
  MP.gtA a (s, t) || (MP.geqA a (s, t) && any (\s' -> geqWPO param s' t) (args st s))
geqWPO param@(a, p, st) s@(F _ _) t@(F _ _) = 
  MP.gtA a (s, t) ||
    (MP.geqA a (s, t) &&
      (any (\s' -> geqWPO param s' t) (args st s) ||
       (all (\t' -> gtWPO param s t') (args st t) &&
        (gtPrec p s t || (geqPrec p s t && geqWPOLex param (args st s) (args st t))))))

gtWPO :: Parameter -> Term -> Term -> Bool
gtWPO (a, _, _) s@(V _) t = MP.gtA a (s, t)
gtWPO param@(a, _, st) s@(F _ _) t@(V _) =
  MP.gtA a (s, t) || (MP.geqA a (s, t) && any (\s' -> geqWPO param s' t) (args st s))
gtWPO param@(a, p, st) s@(F _ _) t@(F _ _) =
  MP.gtA a (s, t) ||
    (MP.geqA a (s, t) &&
      (any (\s' -> geqWPO param s' t) (args st s) ||
       (all (\t' -> gtWPO param s t') (args st t) &&
        (gtPrec p s t || (geqPrec p s t && gtWPOLex param (args st s) (args st t))))))

-- args are assumed to be already filtered
geqWPOLex :: Parameter -> [Term] -> [Term] -> Bool
geqWPOLex _ [] [] = True
geqWPOLex _ [] (_ : _) = False
geqWPOLex _ (_ : _) []  = True
geqWPOLex param (s : ss) (t : ts) =
  gtWPO param s t || 
  (geqWPO param s t && geqWPOLex param ss ts)

gtWPOLex :: Parameter -> [Term] -> [Term] -> Bool
gtWPOLex _ [] [] = False
gtWPOLex _ [] (_ : _) = False
gtWPOLex _ (_ : _) []  = True
gtWPOLex param (s : ss) (t : ts) =
  gtWPO param s t || 
  (geqWPO param s t && gtWPOLex param ss ts)

-- Encoding functions.

type Table = [(Term,Term)]

-- example:
-- f_A(x_1, x_2) = max { a_0_f, b_1_f x_1 + a_1_f, b_2_f x_2 + a_2_f }

varF0 :: String -> Exp
varF0 f = Var ("a_0_" ++ f) 

varFi :: String -> Int -> Exp
varFi f i = Var ("a_" ++ show i ++ "_" ++ f) 

varF'i :: String -> Int -> Formula
varF'i f i = FVar ("b_" ++ show i ++ "_" ++ f) 

varPrec :: String -> Exp
varPrec f = Var ("p_" ++ f)

varStatus :: String -> Int -> Formula
varStatus f i = FVar ("st_" ++ show i ++ "_" ++ f) 

varGeqWPO :: Table -> Rule -> Formula
varGeqWPO table rule
  | Just i <- elemIndex rule table = FVar ("geqWPO_" ++ show i)
  | otherwise                      = error "varGeqWPO"

varGtWPO :: Table -> Rule -> Formula
varGtWPO table rule
  | Just i <- elemIndex rule table = FVar ("gtWPO_" ++ show i)
  | otherwise                      = error "varGtWPO"


symbolicMaxPlus :: Term -> SymbolicMaxPlus
symbolicMaxPlus (V x) = SMaxPlus.var x
symbolicMaxPlus (F f ts) =
   SMaxPlus.flatten (SMaxPlus.const (varF0 f) : [ multiplyAdd f i ti | (i, ti) <- zip [1..] ts ])
  where
    multiplyAdd fname i ti = SMaxPlus.addConstant (varFi fname i) (SMaxPlus.mult (varF'i fname i) (symbolicMaxPlus ti))

-- s >=_A t
encodeGeqA :: Rule -> Formula
encodeGeqA (l, r) = SMaxPlus.encodeGeq (symbolicMaxPlus l) (symbolicMaxPlus r)

-- s >_A t
encodeGtA :: Rule -> Formula
encodeGtA (l, r) = SMaxPlus.encodeGt (symbolicMaxPlus l) (symbolicMaxPlus r)

-- f >= g
encodeGeqPrec :: Rule -> Formula
encodeGeqPrec (F f _, F g _) = SMT.geq (varPrec f) (varPrec g)
encodeGeqPrec _ = error "encodeGeqPrec"

-- f > g
encodeGtPrec :: Rule -> Formula
encodeGtPrec (F f _, F g _) = SMT.gt (varPrec f) (varPrec g)
encodeGtPrec _ = error "encodeGtPrec"

-- s >_wpo t
encodeGtWPO :: Table -> Rule -> Formula
encodeGtWPO _ rl@(V _, _) = encodeGtA rl
encodeGtWPO table rl@(F _ _, V _) =
  disj [ encodeGtA rl, conj [ encodeGeqA rl, encodeWPOSub table rl] ]
encodeGtWPO table rl@(F f ss, F g ts) =
  disj [ encodeGtA rl,
    conj [ encodeGeqA rl,
      disj [
        encodeWPOSub table rl,
        conj [
          encodeWPOArgs table rl,
          disj [
            encodeGtPrec rl,
            -- only strict precedence is considered for now
            if f == g then encodeGtWPOLex table f ss ts else bottom
          ]
        ]
      ]
    ]
  ]


-- s >=_wpo t
encodeGeqWPO :: Table -> Rule -> Formula
encodeGeqWPO _ (V x, V y) = if x == y then top else bottom
encodeGeqWPO _ rl@(V _, _) = encodeGtA rl
encodeGeqWPO table rl@(F _ _, V _) =
  disj [ encodeGtA rl, conj [ encodeGeqA rl, encodeWPOSub table rl] ]
encodeGeqWPO table rl@(F f ss, F g ts) =
  disj [ encodeGtA rl,
    conj [ encodeGeqA rl,
      disj [
        encodeWPOSub table rl,
        conj [
          encodeWPOArgs table rl,
          disj [
            encodeGtPrec rl,
            -- only strict precedence is considered for now
            if f == g then encodeGeqWPOLex table f ss ts else bottom
          ]
        ]
      ]
    ]
  ]

encodeWPOSub :: Table -> Rule -> Formula
encodeWPOSub table (F f ss, t) =
 disj [ conj [ varStatus f i, varGeqWPO table (s', t) ] | (s', i) <- zip ss [1..] ]
encodeWPOSub _ _ = error "encodeWPOSub"

encodeWPOArgs :: Table -> Rule -> Formula
encodeWPOArgs table (s, F g ts) =
 conj [ implies (varStatus g j) (varGtWPO table (s, t')) | (t', j) <- zip ts [1..] ]
encodeWPOArgs _ _ = error "encodeWPOArgs"

encodeGtWPOLex :: Table -> String -> [Term] -> [Term] -> Formula
encodeGtWPOLex tbl f ss ts = encodeGtWPOLex' tbl f 1 ss ts

encodeGtWPOLex' :: Table -> String -> Int -> [Term] -> [Term] -> Formula
encodeGtWPOLex' _ _ _ [] [] = bottom
encodeGtWPOLex' tbl f i (s : ss) (t : ts) =
  disj [ conj [ varStatus f i, varGtWPO tbl (s, t) ],
         conj [ implies (varStatus f i) (varGeqWPO tbl (s, t)), encodeGtWPOLex' tbl f (i+1) ss ts ] ]
encodeGtWPOLex' _ _ _ _ _ = error "length of terms must be the same"

encodeGeqWPOLex :: Table -> String -> [Term] -> [Term] -> Formula
encodeGeqWPOLex tbl f ss ts = encodeGeqWPOLex' tbl f 1 ss ts

encodeGeqWPOLex' :: Table -> String -> Int -> [Term] -> [Term] -> Formula
encodeGeqWPOLex' _ _ _ [] [] = top
encodeGeqWPOLex' tbl f i (s : ss) (t : ts) =
  disj [ conj [ varStatus f i, varGtWPO tbl (s, t) ],
         conj [ implies (varStatus f i) (varGeqWPO tbl (s, t)), encodeGeqWPOLex' tbl f (i+1) ss ts ] ]
encodeGeqWPOLex' _ _ _ _ _ = error "length of terms must be the same"

encode_one :: TRS -> TRS -> Formula
encode_one trs1 trs2 =
  conj (disj [ varGtWPO table rl | rl <- trs1 ] :
        [ varGeqWPO table rl | rl <- trs1 ++ trs2 ] ++
        [ SMT.geq (varF0 f) (Val 0) | (f, _) <- fs ] ++
        [ SMT.geq (varFi f i) (Val 0) | (f, n) <- fs, i <- [1..n] ] ++
         -- pi-simplicity
        [ implies (varStatus f i) (varF'i f i) | (f, n) <- fs, i <- [1..n] ] ++
        -- strict precedence 
        [ Distinct [ varPrec f | (f, _) <- fs ] ] ++
        [ implies (varGtWPO table rule) (encodeGtWPO table rule) | rule <- table ] ++
        [ implies (varGeqWPO table rule) (encodeGeqWPO table rule) | rule <- table ])
  where 
    fs = union (TRS.signatureOf trs1) (TRS.signatureOf trs2)
    table = nub [ (s', t') | (lhs, rhs) <- trs1 ++ trs2,
                             s' <- subterms lhs,
                             t' <- subterms rhs ]

encode_all :: TRS -> TRS -> Formula
encode_all trs1 trs2 =
  conj ([ varGtWPO table rl | rl <- trs1 ] ++
        [ varGeqWPO table rl | rl <- trs1 ++ trs2 ] ++
        [ SMT.geq (varF0 f) (Val 0) | (f, _) <- fs ] ++
        [ SMT.geq (varFi f i) (Val 0) | (f, n) <- fs, i <- [1..n] ] ++
         -- pi-simplicity
        [ implies (varStatus f i) (varF'i f i) | (f, n) <- fs, i <- [1..n] ] ++
        -- strict precedence 
        [ Distinct [ varPrec f | (f, _) <- fs ] ] ++
        [ implies (varGtWPO table rule) (encodeGtWPO table rule) | rule <- table ] ++
        [ implies (varGeqWPO table rule) (encodeGeqWPO table rule) | rule <- table ])
  where 
    fs = union (TRS.signatureOf trs1) (TRS.signatureOf trs2)
    table = nub [ (s', t') | (lhs, rhs) <- trs1 ++ trs2,
                             s' <- subterms lhs,
                             t' <- subterms rhs ]

-- Decoding functions

decodeInterpretation :: String -> Int -> Model -> MP.Interpretation
decodeInterpretation f n model = (f, (f0, ff's))
  where 
    f0 = evalExp model (varF0 f)
    evalf fname i = evalExp model (varFi fname i)
    evalf' fname i = if evalFormula model (varF'i fname i) then 1 else 0
    ff's = [ (evalf f i, evalf' f i) | i <- [1 .. n] ]

decodeAlgebra :: Signature -> Model -> MP.Algebra
decodeAlgebra sig model =
  [ decodeInterpretation f n model | (f, n) <- sig ]

decodePrecedence :: Signature -> Model -> Precedence
decodePrecedence sig model =
  [ (f, evalExp model (varPrec f)) | (f, _) <- sig ]

decodeStatus :: Signature -> Model -> Status
decodeStatus sig model =
  [ (f, [ evalFormula model (varStatus f i) | i <- [1..n] ]) | (f, n) <- sig ]

decode :: TRS -> TRS -> Model -> Parameter
decode r s model =
  (decodeAlgebra sig model,
   decodePrecedence sig model,
   decodeStatus sig model)
  where
    sig =  union (TRS.signatureOf r) (TRS.signatureOf s)

solveOne :: String -> Problem -> IO (Maybe (Order, TRS, Problem))
solveOne tool (ps, rs, m) = do
  result <- sat tool (encode_one ps us)
  case result of
    Nothing -> return Nothing
    Just model 
      | length ps' < length ps ->
          return (Just (MaxPlusWPO a p st, us, (ps', rs, m))) 
      | otherwise -> error "solveOne"
      where 
        params@(a, p, st) = decode ps us model
        ps' = [ rule | rule@(l, r) <- ps, not (gtWPO params l r) ]
  where
    us = Usable.usable_rules ps rs

solveAll :: String -> Problem -> IO (Maybe Order)
solveAll tool (ps, rs, _) = do
  result <- sat tool (encode_all ps rs)
  case result of
    Nothing -> return Nothing
    Just model -> do
          forM_ ps checkGt
          forM_ rs checkGeq
          return (Just (MaxPlusWPO a p st)) 
      where 
        param@(a, p, st) = decode ps rs model
        checkGt (s, t)
          | gtWPO param s t = return ()
          | otherwise = error (show s ++ " > " ++ show t ++ " is not satisfied")
        checkGeq (s, t)
          | geqWPO param s t = return ()
          | otherwise = error (show s ++ " >= " ++ show t ++ " is not satisfied")
